from __future__ import annotations

from enum import Enum
from typing import Callable, Generator, Iterable

import logging
from avariantas.information.datatags import DataTags
from avariantas.information.mutations import (
    CompoundGroupName,
    GroupName,
    OrganizedMutations,
)
import numpy as np
from scipy.stats import chi2_contingency
from avariantas.information.representations import TestTable
from avariantas.problems.exceptions import AssociationTestFailedException

from collections import deque
from numpy import ndarray, prod, sum
from scipy.special import comb, factorial
from scipy.stats import fisher_exact


logger = logging.getLogger(__name__)


class AssociationTest(Enum):
    FISHERS_EXACT = 0
    FISHERS = 1
    CHI_SQUARE = 2
    YATES_CHI_SQUARE = 3

    @staticmethod
    def test_of(name: AssociationTest) -> Callable[[np.ndarray], float]:
        _callable_map = {
            AssociationTest.FISHERS_EXACT: lambda table: fishers_exact_test(
                table, True
            ),
            AssociationTest.FISHERS: lambda table: fishers_exact_test(table, False),
            AssociationTest.CHI_SQUARE: lambda table: chi2_contingency(
                table, correction=False
            ).pvalue,  # type: ignore
            AssociationTest.YATES_CHI_SQUARE: lambda table: chi2_contingency(
                table, correction=True
            ).pvalue,  # type: ignore
        }
        return _callable_map[name]


def group_vs_group_association(
    mutations: OrganizedMutations,
    group_key_a: str,
    group_key_b: str,
    count_individual: bool = False,
    group_a: CompoundGroupName | GroupName | None = None,
    group_b: CompoundGroupName | GroupName | None = None,
    test_priority: list[AssociationTest] | AssociationTest = AssociationTest.CHI_SQUARE,
) -> TestTable:
    """A single association test between two groups of data

    A contingency table is built, and tested for via parameters from
        :class:`AssociationTest`

    :param mutations: The organized mutations to use for the association test
    :type mutations: OrganizedMutations
    :param group_key_a: The group key representing the first category in the
           association test
    :type group_key_a: str
    :param group_key_b: The group key representing the second category in the
           association test
    :type group_key_b: str
    :param count_in_mutation: If we should try to count multiple occurrences within a
           mutation, defaults to True
    :type count_in_mutation: bool
    :param group_a: The group containing the data to be used for the first category,
           defaults to None
    :type group_a: CompoundGroupName | GroupName
          | None, optional
    :param group_b: The group containing the data to be used for the second category,
          defaults to None
    :type group_b: CompoundGroupName | GroupName
          | None, optional
    :param test_priority: The association test to run in the given order,
           defaults to AssociationTest.CHI_SQUARE
    :type test_priority: list[AssociationTest] | AssociationTest, optional
    :raises AssociationTestFailedException: When the resulting contingency table cannot
            be tested with the given tests
    :return: A float (or none if table has bad shape) and a representing table
    :rtype: tuple[float | None, Table]
    """
    TEST_NAME = "ASSOCIATION"
    logger.info(
        'Performing association test between "%s" and "%s".', group_key_a, group_key_b
    )
    if group_a is None:
        group_a = mutations.union_all
    if group_b is None:
        group_b = mutations.union_all
    a_subgroups = [
        mutations.intersect(GroupName(group_key_a, group_values), group_a)
        for group_values in mutations.known_group_values(group_key_a)
    ]
    tag_b_values: list[str | Iterable[str]] = list(
        mutations.known_group_values(group_key_b)
    )
    b_subgroups = [
        mutations.intersect(GroupName(group_key_b, group_values), group_b)
        for group_values in tag_b_values
    ]
    if count_individual:
        unique_tag_b_vals = set()
        for possible_b_value in mutations.known_group_values(group_key_b):
            if not isinstance(possible_b_value, str) and isinstance(
                possible_b_value, Iterable
            ):
                unique_tag_b_vals.update(possible_b_value)
            else:
                unique_tag_b_vals.add(possible_b_value)
        tag_b_values = list(unique_tag_b_vals)
        b_subgroups = []

    initial_table_dims = (
        len(a_subgroups),
        len(b_subgroups) if not count_individual else len(tag_b_values),
    )
    logger.debug(
        "Largest possible contingency table with given keys: %s",
        str(initial_table_dims),
    )
    contingency_table = np.zeros(initial_table_dims)
    row_names = []
    col_names = []
    for a_index, a_group in enumerate(a_subgroups):
        row_names.append(str(a_group))
        for b_index, b_group in enumerate(
            b_subgroups if not count_individual else tag_b_values
        ):
            col_names.append(str(b_group))
            if not isinstance(b_group, str) and not isinstance(b_group, Iterable):
                intersect_group = mutations.intersect(
                    a_group,
                    b_group,
                )
            else:
                intersect_group = mutations.intersect(
                    a_group,
                    group_b,
                )
            contingency_table[a_index, b_index] = mutations.count_occurrences(
                intersect_group,
                group_key_b if count_individual else "",
                str(tag_b_values[b_index]) if count_individual else "",
            )

    # Remove all rows with only 0
    contingency_table = contingency_table[:, ~np.all(contingency_table == 0, axis=0)]
    contingency_table = contingency_table[~np.all(contingency_table == 0, axis=1)]
    width, height = contingency_table.shape
    logger.debug("Contingency table actual size: %s", str((width, height)))
    cont_table_datatags = DataTags(
        # TODO Add datatags here
    )
    params = {
        "group_key_a": group_key_a,
        "group_key_b": group_key_b,
        "count_individual": count_individual,
        "group_a": str(group_a),
        "group_b": str(group_b),
        "test_priority": test_priority.name
        if isinstance(test_priority, AssociationTest)
        else [test.name for test in test_priority],
    }
    if width < 2 or height < 2:
        logger.debug("Skipping contingency table.")
        return TestTable(
            TEST_NAME,
            params,
            row_names,
            col_names,
            contingency_table.tolist(),
            None,
            datatags=cont_table_datatags,
        )

    if isinstance(test_priority, AssociationTest):
        test_priority = [test_priority]

    for test in test_priority:
        logger.debug('Attempting to perform "%s" on contingency table.')
        try:
            p_val = AssociationTest.test_of(test)(contingency_table)
            return TestTable(
                TEST_NAME,
                params,
                row_names,
                col_names,
                contingency_table.tolist(),
                p_val,
                datatags=cont_table_datatags,
            )
        except ValueError as e:
            logger.info('Could not perform "%s" for reason: %s', str(test), str(e))
            pass
    raise AssociationTestFailedException(
        f"None of {test_priority} methods could successfully run association tests on "
        "table of shape {width}x{height}."
    )


def group_vs_group_association_over_group(
    mutations: OrganizedMutations,
    over_group_key: str,
    group_key_a: str,
    group_key_b: str,
    count_in_mutations: bool = False,
    group_a: CompoundGroupName | GroupName | None = None,
    group_b: CompoundGroupName | GroupName | None = None,
    test_priority: list[AssociationTest] | AssociationTest = AssociationTest.CHI_SQUARE,
) -> Generator[TestTable, None, None]:
    """Performs association test between two groups over all values of a group key

    :param mutations: The organized mutations to use
    :type mutations: OrganizedMutations
    :param over_group_key: The group key to perform each association test for
    :type over_group_key: str
    :param group_key_a: The first group key the association test is performed between
    :type group_key_a: str
    :param group_key_b: The second group key the association test is performed between
    :type group_key_b: str
    :param count_in_mutation: If the key of group b has multiple values, count the
           occurrences over all the values for the key for each mutation, defaults to
           True
    :type count_in_mutation: bool
    :param group_a: A specific group to use for the first group, defaults to None
    :type group_a: CompoundGroupName | GroupName
        | None, optional
    :param group_b: A specific group to use for the second group, defaults to None
    :type group_b: CompoundGroupName | GroupName
        | None, optional
    :yield: The p-value of iteration (or none if unsuccessful), and the representing
        table
    :rtype: Generator[tuple[float | None, Table], None, None]
    """
    if group_a is None:
        group_a = mutations.union_all
    if group_b is None:
        group_b = mutations.union_all
    for over_key_value in mutations.known_group_values(over_group_key):
        over_key_value_group_a = mutations.intersect(
            GroupName(over_group_key, over_key_value), group_a
        )
        over_key_value_group_b = mutations.intersect(
            GroupName(over_group_key, over_key_value), group_b
        )
        representation = group_vs_group_association(
            mutations,
            group_key_a,
            group_key_b,
            count_in_mutations,
            over_key_value_group_a,
            over_key_value_group_b,
            test_priority,
        )
        yield representation


def fishers_exact_test(table: ndarray, exact=False) -> float:
    height, width = table.shape
    if height == 2 and width == 2:
        _, pval = fisher_exact(table)
        return pval
    elif height == 2 and width > 2:
        return ffh_association_2xc(table, exact=exact)
    else:
        raise ValueError(
            "Unsupported table dimension %dx%d for Fisher's exact test", width, height
        )


def ffh_association_2xc(table: ndarray, exact=False) -> float:
    """Fisher's Exact test with Freeman-Halton extension on 2xC table

    :param table: The contingency table. Must be 2xC
    :type table: ndarray
    :param exact: Whether or not to use exact calculations, defaults to False
    :type exact: bool, optional
    :raises ValueError: The height of the contingency table must be <= 2
    :return: The calculated p-value
    :rtype: float
    """
    height, width = table.shape
    if height > 2:
        raise ValueError("Height of table cannot be greater than 2")
    col_margins = sum(table, axis=0)
    row_margins = sum(table, axis=1)  # items in each row summed
    product_of_coefficient = prod(
        comb(col_margins, table[0], exact=False)
        if not exact
        else tuple(exact_comb_wrapper(col_margins, table[0]))
    )
    p_val = float(
        sum(
            [
                prod(comb(col_margins, comb_values))
                for comb_values in combinations_for_sum(int(row_margins[0]), width)
                if all(comb_values <= col_margins)
                and prod(comb(col_margins, comb_values)) <= product_of_coefficient
            ]
        )
        / (
            factorial(sum(row_margins), exact=exact)
            / prod(factorial(row_margins, exact=exact))
        )
    )
    return p_val


def exact_comb_wrapper(n: Iterable[float], k: Iterable[float]):
    for n_c, k_c in zip(n, k):
        yield comb(n_c, k_c, exact=True)


def combinations_for_sum(sum: int, num: int) -> Generator[tuple[int, ...], None, None]:
    """Rapidly generates all possible permutations of integers that with an
    exact `sum` comprised of exactly `num` numbers

    :param sum: The sum of all permutations
    :type sum: int
    :param num: The length of the permutations
    :type num: int
    :yield: A tuple of length `num` and sum of `sum`
    :rtype: Generator[tuple[int, ...], None, None]
    """
    work: deque[tuple[None | int, list[int], int, int]] = deque(
        [(None, [0] * num, sum, int(0))]
    )
    while len(work) > 0:
        parent_remainder, current_combination, remainder, index = work.pop()
        if parent_remainder is not None:
            current_combination[index - 1] = parent_remainder - remainder
        if remainder == 0:
            yield tuple(current_combination)
            continue  # If node is complete, don't continue building
        for new_remainder in range(remainder + 1):
            if index + 1 > num:
                continue
            work.append(
                (remainder, list(current_combination), new_remainder, index + 1)
            )
