from __future__ import annotations
from collections import defaultdict
from enum import Enum
import re
from typing import Generator, Iterable
import logging

from avariantas.information.features import Annotation

from avariantas.information.datatags import DataTags
from avariantas.problems.exceptions import (
    OrganizedMutationException,
    GroupNameParsingException,
    GroupNameException,
)


logger = logging.getLogger(__name__)


class Flags(Enum):
    """Flags for mutations"""

    DELETION = "DELETION"
    INSERTION = "INSERTION"
    INDEL = "INDEL"
    MONOMORPHIC = "MONOMORPHIC"
    SNP = "SNP"
    SV = "SV"
    TRANSITION = "TRANSITION"


class KnownTags(Enum):
    POSITION = 0
    LENGTH = 1
    REF_ALLELE = 2
    GENOTYPE = 3
    FLAGS = 5


class Mutation(Annotation):
    """Represents a mutation as a sequence annotation"""

    def __init__(
        self,
        position: int,
        length: int,
        ref_allele: str,
        genotype: Iterable[str],
        flags: Iterable[Flags],
        datatags: DataTags = DataTags(),
    ):
        """Instantiates a mutation annotation

        :param position: The position of the mutation
        :type position: int
        :param length: The length of the mutated region
        :type length: int
        :param ref_allele: The reference allele
        :type ref_allele: str
        :param genotype: The genotype where each allele is an element in the iterable
        :type genotype: Iterable[str]
        :param flags: The mutation flags
        :type flags: tuple[Flags]
        :param datatags: The datatags for this mutation, defaults to Datatags()
        :type datatags: Datatags, optional
        """
        super().__init__(position, length, datatags)
        self.datatags[KnownTags.REF_ALLELE.name] = ref_allele
        self.datatags[KnownTags.GENOTYPE.name] = tuple(genotype)
        self.datatags[KnownTags.FLAGS.name] = tuple([flag.name for flag in flags])

    def has_flag(self, flag: Flags):
        return flag in [
            Flags[flag_name] for flag_name in self.datatags[KnownTags.FLAGS.name]
        ]

    @property
    def ref_allele(self):
        return str(self.datatags[KnownTags.REF_ALLELE.name])

    @property
    def genotype(self):
        return tuple(self.datatags[KnownTags.GENOTYPE.name])

    @property
    def flags(self):
        return tuple([Flags[name] for name in self.datatags[KnownTags.FLAGS.name]])

    @property
    def pretty_genotype(self):
        return "|".join(self.genotype)

    def __hash__(self):
        return hash(self.datatags)

    def __eq__(self, other):
        return self.datatags == other.datatags


class GroupOperator(Enum):
    INTERSECT = "&"
    UNION = "|"

    def symbolize(self):
        return self.value

    @staticmethod
    def from_symbol(symbol: str):
        return GroupOperator(symbol)


class GroupName:
    GROUP_KEY_NAME_REGEX = r"^([\\\w\/\-+_|{}\[\]\.]+)\=([\\\w\/\-+_|{}\[\]\.\,]+)$"
    GROUP_ALIAS_REGEX = r"^\w+$"

    def __init__(
        self, key, value: str | Iterable[str], alias: str | None = None
    ) -> None:
        self._key = key
        self._value = value if isinstance(value, str) else tuple(sorted(value))
        self._alias = alias
        if not (
            re.fullmatch(GroupName.GROUP_KEY_NAME_REGEX, str(self))
            or re.fullmatch(GroupName.GROUP_ALIAS_REGEX, str(self))
        ):
            raise GroupNameException(
                "The group name must satisfy "
                f"{GroupName.GROUP_KEY_NAME_REGEX} where the key is to "
                'the left of the "=" symbol while the value is to the right. '
                f"Provided value currently translates to: {str(self)}. "
                "Additionally, an alias may be provided that must satisfy "
                f'"{GroupName.GROUP_ALIAS_REGEX}".'
            )

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> str | Iterable[str]:
        return self._value

    @property
    def alias(self) -> str | None:
        return self._alias

    def __hash__(self) -> int:
        return hash((self.key, self.value))

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, GroupName)
            and self.key == __value.key
            and self.value == __value.value
        )

    def __str__(self) -> str:
        if self.alias is not None:
            return self.alias
        val = (
            self.value
            if isinstance(self.value, str)
            else "[" + ",".join(self.value) + "]"
        )
        return f"{self.key}={val}"


class CompoundGroupName:
    def __init__(
        self,
        operation: GroupOperator,
        groups: Iterable[CompoundGroupName | GroupName],
        alias: str | None = None,
    ) -> None:
        self._operation = operation
        self._groups = frozenset(groups)
        if alias is not None and not re.fullmatch(GroupName.GROUP_ALIAS_REGEX, alias):
            raise GroupNameException(
                "An alias must satisfy the following regex "
                f'"{GroupName.GROUP_ALIAS_REGEX}" (provided: "{alias}").'
            )
        self._alias = alias

    @property
    def operation(self):
        return self._operation

    @property
    def groups(self):
        return self._groups

    @property
    def alias(self) -> str | None:
        return self._alias

    def __hash__(self) -> int:
        return hash((self.operation, self.groups))

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, CompoundGroupName)
            and self.operation == __value.operation
            and self.groups == __value.groups
        )

    def __str__(self) -> str:
        if self.alias:
            return self.alias
        symbol = self.operation.symbolize()
        stringified = "("
        for group in self.groups:
            stringified += str(group) + symbol
        stringified = stringified[:-1]
        stringified += ")"
        return stringified


class OrganizedMutations:
    """All mutations organized into groups where each group is a group key, and a group
    key value.

    These groups can then be intersected or unioned to produce compound groups.
    :raises DuplicateMutationException: If two mutations with identical data is
            submitted
    :raises IllegalStateChangeException: If an illegal operation was attempted
    """

    def __init__(self, mutations: Iterable[Mutation] = []) -> None:
        self._mutations: set[Mutation] = set()
        self._compound_group_names: set[CompoundGroupName] = set()
        self._primitive_group_names: set[GroupName] = set()
        self._aliases: dict[str, (CompoundGroupName | GroupName)] = {}
        self._groups: dict[
            (CompoundGroupName | GroupName),
            set[Mutation],
        ] = defaultdict(set)
        self._known_group_keys: dict[
            str,
            set[str | Iterable[str]],
        ] = defaultdict(set)
        self.add_all(mutations)
        self._permanent_groups: tuple[CompoundGroupName | GroupName] = tuple(
            self._groups.keys()
        )
        if logger.level >= logging.DEBUG:
            logger.debug(
                "A list of unique known groups are as follows:\ngroup \tvalue\n"
            )
            for key, value in self._known_group_keys.items():
                logger.debug("%s\t%s", key, value)

    @property
    def mutations(self):
        return frozenset(self._mutations)

    @property
    def all_group_names(
        self,
    ) -> tuple[CompoundGroupName | GroupName]:
        return tuple(self._groups.keys())

    @property
    def all_compound_groups(self) -> tuple[CompoundGroupName]:
        return tuple(self._compound_group_names)

    @property
    def all_primitive_groups(self) -> tuple[GroupName]:
        return tuple(self._primitive_group_names)

    @property
    def union_all(self) -> CompoundGroupName:
        return self.union(alias="all")

    def resolve_alias(self, alias: CompoundGroupName | GroupName | str):
        return self._aliases[alias] if isinstance(alias, str) else alias

    def resolve_all_aliases(
        self, *group_names: CompoundGroupName | GroupName | str
    ) -> tuple[CompoundGroupName | GroupName]:
        results = []
        for group_name in group_names:
            if isinstance(group_name, str):
                results.append(self.resolve_alias(group_name))
            else:
                results.append(group_name)
        return tuple(results)

    def intersect(
        self,
        *group_names: CompoundGroupName | GroupName | str,
        alias: str | None = None,
    ) -> CompoundGroupName:
        if len(group_names) < 1:
            group_names = self.all_group_names
        resolved_group_names = self.resolve_all_aliases(*group_names)
        if len(resolved_group_names) < 2:
            raise OrganizedMutationException(
                f"Cannot intersect less than two groups ({resolved_group_names})."
            )
        intersect_name = CompoundGroupName(
            GroupOperator.INTERSECT, resolved_group_names, alias=alias
        )
        if intersect_name in self._groups:
            return intersect_name
        if alias in self._aliases:
            return self.resolve_alias(alias)  # type: ignore
        built_groups = list(self.build_group_from_name(*resolved_group_names))
        for group in built_groups:
            if group not in self._groups:
                raise GroupNameException(
                    f'"{group}" does not exist in this organized mutation group.'
                )
        self._compound_group_names.add(intersect_name)
        self._groups[intersect_name] = self._groups[built_groups[0]].intersection(
            *[self._groups[group] for group in built_groups[1:]]
        )
        if alias is not None:
            self._aliases[alias] = intersect_name
        return intersect_name

    def union(
        self,
        *group_names: CompoundGroupName | GroupName | str,
        alias: str | None = None,
    ) -> CompoundGroupName:
        if len(group_names) < 1:
            group_names = self.all_group_names
        resolved_group_names = self.resolve_all_aliases(*group_names)
        if len(resolved_group_names) < 2:
            raise OrganizedMutationException(
                f"Cannot union less than two groups ({resolved_group_names})."
            )
        union_name = CompoundGroupName(
            GroupOperator.UNION, resolved_group_names, alias=alias
        )
        if union_name in self._groups:
            return union_name
        if alias in self._aliases:
            return self.resolve_alias(alias)  # type: ignore
        built_groups = list(self.build_group_from_name(*resolved_group_names))
        for group in built_groups:
            if group not in self._groups:
                raise GroupNameException(
                    f'"{group}" does not exist in this organized mutation group.'
                )
        self._compound_group_names.add(union_name)
        self._groups[union_name] = self._groups[built_groups[0]].union(
            *[self._groups[group] for group in built_groups[1:]]
        )
        if alias is not None:
            self._aliases[alias] = union_name
        return union_name

    def build_group_from_name(
        self, *groups: CompoundGroupName | GroupName
    ) -> Generator[CompoundGroupName | GroupName, None, None]:
        for group in groups:
            if group in self._groups:
                yield group
                continue
            if isinstance(group, CompoundGroupName):
                if group.operation == GroupOperator.INTERSECT:
                    yield self.intersect(*group.groups)
                if group.operation == GroupOperator.UNION:
                    yield self.union(*group.groups)
            elif isinstance(group, GroupName):
                yield group

    def add(self, mutation: Mutation) -> None:
        if mutation in self._mutations:
            raise OrganizedMutationException("Cannot have two identical mutations")
        self._mutations.add(mutation)
        for datatags_key in mutation.datatags.all:
            datatags_value = mutation.datatags[datatags_key]
            datatags_group_name = GroupName(datatags_key, datatags_value)
            self._primitive_group_names.add(datatags_group_name)
            self._groups[datatags_group_name].add(mutation)
            self._known_group_keys[datatags_key].add(datatags_value)

    def add_all(self, mutations: Iterable[Mutation]) -> None:
        logger.info("Cataloguing mutations. This may take a while...")
        newly_catalogued = 0
        for mutation in mutations:
            self.add(mutation)
            newly_catalogued += 1
        logger.info(
            "A total of %d mutations were successfully catalogued. There are now %d "
            "groups (%d unique group keys).",
            newly_catalogued,
            len(self.all_group_names),
            len(self.known_group_keys),
        )

    def count_occurrences(
        self,
        group_name: GroupName | CompoundGroupName | str,
        count_tag_key: str = "",
        count_tag_value: str = "",
    ) -> int:
        if bool(count_tag_key) != bool(count_tag_value):
            raise ValueError(
                "Both group key and values are required if either are to be provided."
            )
        group_name = self.resolve_alias(group_name)
        self.build_group_from_name(group_name)
        count = 0
        if len(count_tag_key) > 0 and len(count_tag_value) > 0:
            for mutation in self.mutations_for_group(group_name):
                if count_tag_key not in mutation.datatags:
                    continue
                tag_values = mutation.datatags[count_tag_key]
                if not isinstance(tag_values, str):
                    for tag_value in tag_values:
                        if tag_value == count_tag_value:
                            count += 1
                else:
                    if tag_values == count_tag_value:
                        count += 1
            return count
        return len(self._groups[group_name])

    @property
    def known_group_keys(self) -> frozenset[str]:
        return frozenset(self._known_group_keys.keys())

    def all_aliases(self):
        return tuple(self._aliases.keys())

    def known_group_values(self, key: str) -> frozenset[str | Iterable[str]]:
        if key not in self._known_group_keys:
            raise ValueError(f'"{key}" was not found in known group keys.')
        return frozenset(self._known_group_keys[key])

    def mutations_for_group(
        self,
        group: CompoundGroupName | GroupName | str,
    ) -> frozenset[Mutation]:
        group = self.resolve_alias(group)
        self.build_group_from_name(group)
        return frozenset(self._groups[group])

    def subgroups_of_group(
        self,
        group: CompoundGroupName | GroupName | str,
    ) -> tuple[CompoundGroupName | GroupName]:
        group = self.resolve_alias(group)
        if isinstance(group, CompoundGroupName):
            return tuple(group.groups)
        elif isinstance(group, GroupName):
            return tuple([group])


def group_name_from_query(
    query: str, alias: str | None = None
) -> CompoundGroupName | GroupName:
    def _build_groupname(
        query: str, query_start: int, query_end: int
    ) -> CompoundGroupName | GroupName:
        if query_end is None:
            query_end = len(query)

        root = False
        if query_start == 0 and query_end == len(query):
            root = True

        depth = 0
        union_operations = None
        intersect_operations = None
        par_begin = None
        par_end = None
        for i, c in enumerate(query[query_start:query_end]):
            if c == "(":
                depth += 1
                if par_begin is None:
                    par_begin = i + query_start
            if depth > 0:
                if c == ")":
                    depth -= 1
                    if depth == 0 and par_end is None:
                        par_end = i + query_start
            else:
                if c == "|":
                    union_operations = i + query_start
                elif c == "&":
                    intersect_operations = i + query_start

        # Handles if there are redundant parenthesis around query
        if par_begin == query_start and par_end == query_end - 1:
            return _build_groupname(query, query_start + 1, query_end - 1)

        # Handles the compound group names
        selected_operation_index = None
        selected_operation = None
        if union_operations is not None:
            selected_operation_index = union_operations
            selected_operation = GroupOperator.UNION
        elif intersect_operations is not None:
            selected_operation_index = intersect_operations
            selected_operation = GroupOperator.INTERSECT

        if selected_operation_index is not None and selected_operation is not None:
            return CompoundGroupName(
                selected_operation,
                frozenset(
                    (
                        _build_groupname(
                            query,
                            query_start,
                            selected_operation_index,
                        ),
                        _build_groupname(
                            query,
                            selected_operation_index + 1,
                            query_end,
                        ),
                    )
                ),
                alias=alias if root else None,
            )
        else:
            group_name_match = re.match(
                GroupName.GROUP_KEY_NAME_REGEX, query[query_start:query_end]
            )
            if group_name_match:
                value = group_name_match.group(2)
                return GroupName(
                    group_name_match.group(1),
                    value if "," not in value else value.split(","),
                    alias=alias if root else None,
                )
            else:
                raise GroupNameParsingException(
                    f"Unable to parse query {query}. Invalid syntax."
                )

    return _build_groupname(query, 0, len(query))
