from __future__ import annotations
from typing import Any, Iterable

from avariantas.information.datatags import DataTags
from avariantas.problems.exceptions import RepresentationException


class RepresentationName:
    def __init__(
        self, test_type: str, parameters: Iterable[RepresentationParameter]
    ) -> None:
        self._test_type = test_type
        self._parameters = tuple(parameters)

    @property
    def test_type(self):
        return self._test_type

    @property
    def parameters(self):
        return self._parameters

    def __hash__(self) -> int:
        return hash((self.test_type, self.parameters))

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, RepresentationName)
            and self.test_type == __value.test_type
            and self.parameters == __value.parameters
        )

    def __str__(self) -> str:
        return f"{self.test_type}/{self.parameters}"


class RepresentationParameter:
    def __init__(self, key: str, value: str) -> None:
        self._key = key
        self._value = value

    @property
    def key(self):
        return self._key

    @property
    def value(self):
        return self._value

    def __hash__(self) -> int:
        return hash((self.key, self.value))

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, RepresentationParameter)
            and self.key == __value.key
            and self.value == __value.value
        )

    @staticmethod
    def from_dict(dictionary: dict):
        for key, value in dictionary.items():
            yield RepresentationParameter(key, str(value))

    def __str__(self) -> str:
        return f"{self.key}={self.value}"


class RepresentationResult:
    def __init__(self, name: str, unit: str, value: str) -> None:
        self._name = name
        self._unit = unit
        self._value = value

    @property
    def name(self):
        return self._name

    @property
    def unit(self):
        return self._unit

    @property
    def value(self):
        return self._value

    def __hash__(self) -> int:
        return hash((self.name, self.unit, self.value))

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, RepresentationResult)
            and self.name == __value.name
            and self.unit == __value.unit
            and self.value == __value.value
        )

    def __str__(self) -> str:
        return f"{self._name}: {self.value} {self.unit}"


class Representation:
    def __init__(
        self,
        test_type: str,
        test_params: dict[str, str],
        results: Iterable[RepresentationResult],
        datatags: DataTags = DataTags(),
    ) -> None:
        self._test_type = test_type
        self._test_params = list(RepresentationParameter.from_dict(test_params))
        self._results: dict[str, RepresentationResult] = {}
        for result in results:
            if result.name in self._results:
                raise RepresentationException(
                    f"Two results with the same names cannot exist within one result "
                    f'representation ("{result.name}").'
                )
            self._results[result.name] = result
        self._datatags = datatags

    @property
    def test_type(self):
        return self._test_type

    @property
    def params(self):
        return tuple(self._test_params)

    def result_of(self, name: str):
        return self._results[name]

    @property
    def results(self):
        return tuple(self._results.values())

    @property
    def datatags(self):
        return self._datatags

    def __hash__(self) -> int:
        return hash((self.test_type, self.params, self.datatags))

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, Representation)
            and self.test_type == __value.test_type
            and self.params == __value.params
            and self.datatags == __value.datatags
        )


class OrganizedRepresentations:
    def __init__(self, representations: Iterable[Representation] = []) -> None:
        self._representations: dict[RepresentationName, Representation] = {}

        self.add_all_representation(representations)

    def add_representation(
        self, representation: Representation, ignore_existing: bool = False
    ):
        representation_name = RepresentationName(
            representation.test_type, representation.params
        )
        if representation_name in self._representations:
            if not ignore_existing:
                raise Exception(
                    "Can't insert representation that already exists with the same set "
                    "of parameters."
                )
        self._representations[representation_name] = representation

    def add_all_representation(
        self, representations: Iterable[Representation], ignore_existing: bool = False
    ):
        for representation in representations:
            self.add_representation(representation, ignore_existing)

    def get_representation(
        self, test_type: str, params: Iterable[RepresentationParameter]
    ):
        representation_name = RepresentationName(test_type, params)
        return self._representations[representation_name]

    @property
    def known_representations(self):
        return frozenset(self._representations.keys())


class TestTable(Representation):
    def __init__(
        self,
        test_type: str,
        test_params: dict[str, str],
        row_names: Iterable[str],
        column_names: Iterable[str],
        table: Iterable[Iterable[Any]],
        p_val: float | None,
        datatags: DataTags = DataTags(),
    ) -> None:
        self._row_names = tuple(row_names)
        self._column_names = tuple(column_names)
        self._table = []
        for row in table:
            row_values = []
            for item in row:
                row_values.append(item)
            self._table.append(tuple(row_values))
        self._table = tuple(self._table)
        super().__init__(
            test_type,
            test_params,
            [RepresentationResult("p-val", "N/A", str(p_val))],
            datatags,
        )

    @property
    def row_names(self):
        return self._row_names

    @property
    def column_names(self):
        return self._column_names

    @property
    def table(self):
        return self._table
