from collections import defaultdict, deque
import csv
from glob import glob
import logging
import os
import pickle
import re
from Bio import GenBank
from typing import Generator, Iterable, Union
import vcf
from avariantas.information.genes import Gene

from avariantas.information.mutations import DataTags, Mutation, Flags
from avariantas.information.project import Project
from avariantas.information.utrs import UntranslatedRegion
from avariantas.problems.exceptions import InvalidFileException

logger = logging.getLogger(__name__)


def load_mutations_from_vcf(
    vcf_path: str, additional_datatags: dict[str, str] = {"ingress": "local"}
) -> Generator[Mutation, None, None]:
    """Loads all mutations represented by the VCF

    :param vcf_path: The path to the VCF to load mutations from
    :type vcf_path: str
    :param additional_datatags: Any additional datatags (overwrites),
        defaults to {"ingress": "local"}
    :type additional_datatags: dict[str, str], optional
    :raises InvalidFileException: If the file path does not end in '.vcf'
    :yield: Every :obj:`avariantas.information.mutations.Mutation` that was produced
    :rtype: Generator[Mutation, None, None]
    """

    if os.path.splitext(vcf_path)[1].lower() != ".vcf":
        raise InvalidFileException('Path does not end in ".vcf" file extension')
    with open(vcf_path, "r") as item_fd:
        vcf_reader = vcf.Reader(item_fd)
        for record in vcf_reader:
            gt = ["N/A"]
            if len(record.samples) < 1:
                mutation_datatags = DataTags(
                    CHROM=record.CHROM,
                    QUAL=str(record.QUAL),
                    FILTER=list(record.FILTER),  # type: ignore
                    **(record.INFO | additional_datatags),
                )
                yield Mutation(
                    record.POS,
                    int(record.INFO["LEN"])
                    if "LEN" in record.INFO
                    else 1,  # Hopefully not bad
                    record.REF,
                    gt,
                    tuple(),
                    mutation_datatags,
                )
                continue
            for sample in record.samples:
                mutation_datatags = DataTags(
                    CHROM=record.CHROM,
                    QUAL=str(record.QUAL),
                    FILTER=list(record.FILTER),  # type: ignore
                    sample=str(sample.sample),
                    **(
                        record.INFO
                        | (sample.data._asdict())  # type: ignore
                        | additional_datatags
                    ),
                )
                if sample.gt_bases is not None:
                    gt = sample.gt_bases.split("/")

                flags = []
                if record.is_deletion:
                    flags.append(Flags.DELETION)
                if record.is_indel:
                    flags.append(Flags.INDEL)
                if record.is_monomorphic:
                    flags.append(Flags.MONOMORPHIC)
                if record.is_snp:
                    flags.append(Flags.SNP)
                if record.is_sv:
                    flags.append(Flags.SV)
                if record.is_transition:
                    flags.append(Flags.TRANSITION)

                yield Mutation(
                    record.POS,
                    int(record.INFO["LEN"]) if "LEN" in record.INFO else 0,
                    record.REF,
                    gt,
                    tuple(flags),
                    mutation_datatags,
                )


def load_mutations_from_vcfs(
    vcf_items: Union[Iterable[str], str],
    additional_datatags_csv: str = "",
    datatags_filename_header="",
    recursive: bool = False,
) -> Generator[Mutation, None, None]:
    """Loads mutations from multiple VCF files inside a directory.

    Calls :func:`avariantas.ingress.local.load_mutations_from_vcf` on each item in
    the directory.

    Will expand glob notation when needed.

    A CSV may be provided allowing for additional datatags for each mutation to be added
    by the file. A datatags filename header value provided will indicate which column
    header contains the filenames of the VCF files each row of information should be
    associated with. An empty string (default) will cause the program to use the first
    column in the CSV. This will also override any datatags loaded for the VCF itself,
    as well as the "origin" property which defaults to the folder the VCF was found in.

    :param vcf_items: Path to the directory containing the VCFs to load
    :type vcf_items: str
    :param additional_datatags_csv: Path to a CSV where one column is the file name,
           and the remaining columns further describe the variants for each file,
           defaults to ""
    :type additional_datatags_csv: str, optional
    :param datatags_filename_header: The filename header in the CSV, defaults to ""
    :type datatags_filename_header: str, optional
    :param recursive: Whether or not this operation should be recursive,
           defaults to False
    :type recursive: bool, optional
    :raises ValueError: If the filename header for the additional datatags is provided
            but no path to the additional datatags is provided.
    :yield: All mutations from all given VCFs
    :rtype: Generator[Mutation, None, None]
    """

    if datatags_filename_header and not additional_datatags_csv:
        raise ValueError(
            "datatags_filename_header provided but no additional_datatags_csv provided."
        )

    indices_obtained = False
    filename_index = 0
    other_indices = {}
    additional_datatags = defaultdict(dict)
    if additional_datatags_csv:
        with open(additional_datatags_csv, "r") as csv_fd:
            reader = csv.reader(csv_fd)
            for row in reader:
                if not indices_obtained:
                    if datatags_filename_header:
                        filename_index = row.index(datatags_filename_header)
                    for index, item in enumerate(row):
                        if index == filename_index:
                            continue
                        other_indices[index] = item
                    indices_obtained = True
                    continue
                row_additional_datatags = additional_datatags[row[filename_index]]
                for index, name in other_indices.items():
                    row_additional_datatags[name] = row[index]

    work: deque[str] = deque()
    if isinstance(vcf_items, str):
        vcf_items = [vcf_items]
    for vcf_item in vcf_items:
        if not os.path.isdir(vcf_item):
            globbed = glob(vcf_item, recursive=recursive)
            if len(globbed) == 0:
                logger.warning('No valid VCF files found at "%s".', vcf_item)
            work.extend(globbed)
        else:
            work.extend(
                [
                    os.path.join(vcf_item, vcf_item_base)
                    for vcf_item_base in os.listdir(vcf_item)
                ]
            )
    while len(work):
        item_path = work.pop()
        dir_item = os.path.basename(item_path)
        if recursive and os.path.isdir(item_path):
            work.extend(
                [
                    os.path.join(item_path, nested_item)
                    for nested_item in os.listdir(item_path)
                ]
            )
        elif (
            os.path.isfile(item_path)
            and os.path.splitext(item_path)[1].lower() == ".vcf"
        ):
            curr_additional_datatags = (
                additional_datatags[dir_item] if dir_item in additional_datatags else {}
            )
            if "origin" not in curr_additional_datatags:
                curr_additional_datatags["origin"] = os.path.abspath(item_path)
            yield from load_mutations_from_vcf(item_path, curr_additional_datatags)


def load_genes_from_genbank_file(
    path: str, accession: str
) -> Generator[Gene, None, None]:
    """Loads all genes from a GenBank file with the given accession

    :param path: The path to the GenBank file to load from
    :type path: str
    :param accession: The accession of the record in the GenBank file
    :type accession: str
    :yield: Every :obj:`avariantas.information.genes.Gene` produced
    :rtype: Generator[Gene, None, None]
    """
    with open(path, "r") as genbank_fd:
        for record in GenBank.parse(genbank_fd):
            if record.accession != accession:  # type: ignore
                continue
            for genbank_feature in record:
                f_key = genbank_feature.key  # type: ignore
                if f_key == "gene":
                    start, end = genbank_feature.location.split("..")  # type: ignore
                    start = int(start)
                    end = int(end)
                    gene_name = "N/A"
                    for qualifier in genbank_feature.qualifiers:  # type: ignore
                        if qualifier.key == "/gene=":
                            gene_name = qualifier.value
                            break
                    yield Gene(gene_name, start, end - start + 1)


def load_utrs_from_genbank_file(
    path: str, accession: str
) -> Generator[UntranslatedRegion, None, None]:
    """Loads all UTRs from a GenBank file with the given accession

    :param path: The path to the GenBank file to load from
    :type path: str
    :param accession: The accession of the record in the GenBank file
    :type accession: str
    :yield: Every :obj:`avariantas.information.utrs.UntranslatedRegion` produced
    :rtype: Generator[UntranslatedRegion, None, None]
    """
    with open(path, "r") as genbank_fd:
        for record in GenBank.parse(genbank_fd):
            if not (
                accession in record.accession  # type: ignore
                or accession.split(".")[0] in record.accession  # type: ignore
            ):
                continue
            for genbank_feature in record.features:  # type: ignore
                re_match = re.match(r"(\d)' ?UTR", genbank_feature.key)
                if re_match:
                    prime = int(re_match.group(1))
                    start, end = genbank_feature.location.split("..")
                    start = int(start)
                    end = int(end)
                    yield UntranslatedRegion(prime, start, end - start + 1)


def load_project(file_path: str) -> Project:
    if os.path.splitext(file_path)[1] != ".ava":
        file_path += ".ava"
    logger.debug("Loading from: %s", file_path)
    file_stat = os.stat(file_path)
    warn_filesize = 1024 * 1024 * 16
    if file_stat.st_size > warn_filesize:
        logger.info(
            "Loading project data may take a moment due to large file size (%.2f MB). "
            "Please be patient.",
            file_stat.st_size / (1024 * 1024),
        )
    with open(file_path, "rb") as project_stream:
        project = pickle.load(project_stream, encoding="bytes")
        logger.info('Loaded project data from "%s".', file_path)
        return project
