import logging
from typing import Iterable
from Bio import Entrez
import requests
from avariantas.information.datatags import DataTags

from avariantas.information.mutations import Flags, Mutation

from cachier import cachier

from avariantas.problems.exceptions import IngressException

logger = logging.getLogger(__name__)


@cachier()
def search_dbsnp_rsid(chromosome: str, position: str, max_results: int = 5) -> str:
    """Searches for a rsID by a given chromosome and mutation in NCBI dbSNP

    Uses the dbSNP database from NCBI. Returns "N/A" if no valid results were found.

    :param chromosome: The human chromosome (e.g. X or, 1)
    :type chromosome: str
    :param position: The position of the mutation in the chromosome sequence
    :type position: str
    :param max_results: How many results per request should be fetched, defaults to 5
    :type max_results: int, optional
    :return: The rsid with the "rs" prefix
    :rtype: str
    """
    search_query = f"{position}[POSITION] AND {chromosome}[CHR]"
    logger.debug("Searching SNP database with query: %s", search_query)
    search_handle = Entrez.esearch(db="snp", retmax=max_results, term=search_query)
    result_ids = Entrez.read(search_handle)["IdList"]  # type: ignore
    search_handle.close()
    logger.debug("%d/%d results returned", len(result_ids), max_results)
    for i, result_id in enumerate(result_ids):
        logger.debug("Checking result #%d", i)
        summary_handle = Entrez.esummary(db="snp", id=result_id)
        e_parsed = Entrez.read(summary_handle)
        if len(e_parsed["DocumentSummarySet"]["DocumentSummary"]) > 1:  # type: ignore
            logger.debug(
                "Unexpected summary length %d > 1",
                len(e_parsed["DocumentSummarySet"]["DocumentSummary"]),  # type: ignore
            )
        esummary = e_parsed["DocumentSummarySet"]["DocumentSummary"][0]  # type: ignore

        if esummary["SNP_CLASS"].lower() != "snv":
            logger.warning(
                'Non SNV result found in SNP database with result ID "%s". This will '
                "be skipped.",
                result_id,
            )

        if str(chromosome) != esummary["CHR"] or (
            f"{chromosome}:{position}" != esummary["CHRPOS"]
            and f"{chromosome}:{position}" != esummary["CHRPOS_PREV_ASSM"]
        ):
            logger.debug(
                "Given information does not match with result id %s from NCBI",
                result_id,
            )
            continue
        return "rs" + esummary["SNP_ID"]
    logger.warning(
        'No SNV found with position %d on chromosome "%s".', int(position), chromosome
    )
    return "N/A"


def annotate_snvs_with_rsid(
    mutations: Iterable[Mutation],
) -> tuple[tuple[Mutation], frozenset[str]]:
    """Annotates a given list of mutations with their respective dbSNP rsIDs.

    Only applies to mutations without "rsid", and with "CHROM" as datatags tags.
    Additionally, the mutation must be of length 1 (SNP).

    :param mutations: The mutations to operate on.
    :type mutations: tuple[tuple[Mutation], frozenset[str]]
    """
    mutations = []
    found_rsids: set[str] = set()
    for mutation in mutations:
        if "rsid" in mutation.datatags.all:
            logger.warning(
                "Skipping mutation at %d with length %d due to already having an rsid "
                "datatags group key.",
                mutation.start,
                mutation.length,
            )
            continue
        if "CHROM" not in mutation.datatags:
            logger.warning(
                'Skipping mutation at %d with length %d due to not having a "CHROM" '
                "datatags group key.",
                mutation.start,
                mutation.length,
            )
            continue
        if mutation.length != 1:
            logger.warning(
                "Skipping mutation at %d with length %d due to length being not equal "
                "to 1."
            )
            continue
        rsid = search_dbsnp_rsid(
            mutation.datatags["CHROM"], str(mutation.start)  # type: ignore
        )
        mutation.datatags["rsid"] = rsid
        if rsid != "N/A":
            found_rsids.add(rsid)
        mutations.append(mutation)
    return tuple(mutations), frozenset(found_rsids)


@cachier()
def fetch_ensembl_population_patients(
    population: str,
) -> Iterable[tuple[str, str]]:
    """Fetches patient data for population from Ensembl

    :param population: The population code (e.g. 1000GENOMES:phase_3:GBR)
    :type population: str
    :return: An iterable of all patient IDs and their respective sexes
    :rtype: Iterable[tuple[str, str]]
    """
    results = []
    population_data_url = (
        "https://rest.ensembl.org/info/variation/populations/human/{0}?"
        "content-type=application/json"
    ).format(population.upper())
    population_response = requests.get(population_data_url).json()
    logger.debug("Attempting to request %s population sample data", population)
    if len(population_response) > 1:
        logger.warning(
            'Results from Ensembl contains unused data in population "%s"',
            population_data_url,
        )
    for pop_sample in population_response[0]["individuals"]:
        sample_name = str(pop_sample["name"])
        sex = str(pop_sample["gender"]).lower()
        results.append((sample_name, sex))
    return results


@cachier()
def fetch_ensembl_rsid_population_genotypes(
    rsid: str,
) -> Iterable[Mutation]:
    """Fetches samples for all populations with a given rsID SNP

    :param rsid: The rsID to perform the fetch with
    :type rsid: str
    :return: The sample ID, a list representing the genotype, and the patient sex
    :rtype: Iterable[Mutation]
    """
    results = []
    variant_data_url = (
        "https://rest.ensembl.org/variation/human/{0}"
        "?content-type=application/json;pops=1;genotypes=1"
    ).format(rsid)
    logger.debug("Making GET request to: %s", variant_data_url)
    variant_response = requests.get(variant_data_url.format(rsid=rsid)).json()
    logger.debug("Response obtained from get request.")
    variant_type = str(variant_response["var_class"]).upper()
    for genotype_data in variant_response["genotypes"]:
        sample = str(genotype_data["sample"])
        genotype = tuple(sorted(genotype_data["genotype"].split("|")))
        sex = str(genotype_data["gender"]).lower()
        start = None
        end = None
        ancestral_allele = None
        if "mappings" not in variant_response:
            raise IngressException(
                "No mapping information was obtained from fetched data"
            )
        for mapping in variant_response["mappings"]:
            if mapping["coord_system"] == "chromosome":
                start = int(mapping["start"])
                end = int(mapping["end"])
                ancestral_allele = str(mapping["ancestral_allele"])
        if start is None or end is None or ancestral_allele is None:
            raise IngressException(
                "Start, end, and/or ancestral allele of obtained genotype was none"
            )
        results.append(
            Mutation(
                start,
                end - start + 1,
                ancestral_allele,
                genotype,
                tuple([Flags(variant_type)]),
                DataTags(
                    sample_id=sample, patient_sex=sex, rsid=rsid, ingress="remote"
                ),
            )
        )
    return results


@cachier()
def fetch_genotypes_of_rsid_from_population(
    rsid: str, population: str
) -> Iterable[Mutation]:
    results: list[Mutation] = []
    """Fetches genotypes for a given population of a given SNP defined by a rsID.

    :param rsid: The rsID of the genotypes
    :type rsid: str
    :param population: The population to retrieve genotypes for
    :type population: str
    :return: All samples and their genotypes of the rsID from the defined population
    :rtype: Generator[Mutation, None, None]
    """
    samples_from_population = set()
    for sample_id, _ in fetch_ensembl_population_patients(population):
        samples_from_population.add(sample_id)

    for mutation in fetch_ensembl_rsid_population_genotypes(rsid):
        if mutation.datatags["sample_id"] in samples_from_population:
            results.append(mutation)
    return tuple(results)


def fetch_ensembl_rsids_population_genotypes(rsids: Iterable[str]):
    for rsid in rsids:
        yield from fetch_ensembl_rsid_population_genotypes(rsid)


def fetch_genotypes_of_rsids_from_population(rsids: Iterable[str], population: str):
    for rsid in rsids:
        yield from fetch_genotypes_of_rsid_from_population(rsid, population)


def annotate_mutations_from_ensembl(
    to_annotate: Iterable[Mutation],
    population_code: str = "",
) -> Iterable[Mutation]:
    _, rsids = annotate_snvs_with_rsid(to_annotate)
    if len(population_code) == 0:
        return fetch_ensembl_rsids_population_genotypes(rsids)
    return fetch_genotypes_of_rsids_from_population(rsids, population_code)
