from __future__ import annotations

from argparse import _SubParsersAction, ArgumentParser

from avariantas.analysis.association import (
    AssociationTest,
    group_vs_group_association_over_group,
    group_vs_group_association,
)
from avariantas.egress.local import save_project
from avariantas.ingress.local import load_project


def perform_association_test(
    project_path: str,
    tag_a: str,
    tag_b: str,
    count_individual: bool,
    group_a_name: str,
    group_b_name: str,
    test_type_names: str | list[str],
    repeat_for: str,
):
    project = load_project(project_path)
    organized_mutations = project.organized_mutations
    organized_representations = project.organized_representations
    group_a = (
        organized_mutations.group_name_from_query(group_a_name)
        if group_a_name
        else None
    )
    group_b = (
        organized_mutations.group_name_from_query(group_b_name)
        if group_b_name
        else None
    )
    if isinstance(test_type_names, str):
        test_type_names = [test_type_names]
    test_types = [AssociationTest[test_type.upper()] for test_type in test_type_names]
    results = None
    if repeat_for:
        results = group_vs_group_association_over_group(
            organized_mutations,
            repeat_for,
            tag_a,
            tag_b,
            count_individual,
            group_a,
            group_b,
            test_types,
        )
        organized_representations.add_all_representation(results)
    else:
        results = group_vs_group_association(
            organized_mutations,
            tag_a,
            tag_b,
            count_individual,
            group_a,
            group_b,
            test_types,
        )
        organized_representations.add_representation(results)
    save_project(project, project_path)


def add_associations_parser(subparsers: _SubParsersAction[ArgumentParser]):
    parser = subparsers.add_parser(
        name="association", help="Perform a form of association test."
    )
    parser.add_argument(
        "tag_a",
        type=str,
        help="The tag for the mutation count of the first category.",
        metavar="a",
    )
    parser.add_argument(
        "tag_b",
        type=str,
        help="The tag for the mutation count of the second category.",
        metavar="b",
    )
    parser.add_argument(
        "-c",
        "--count-individuals",
        action="store_true",
        help="Instead of counting an entire mutation as one sample, "
        "count all the individual multiple values of a group name "
        "for each mutation as a sample.",
        default=False,
        required=False,
    )
    parser.add_argument("-A", "--group-a", type=str, required=False, default="")
    parser.add_argument("-B", "--group-b", type=str, required=False, default="")
    parser.add_argument(
        "-t",
        "--test-type",
        type=str,
        nargs="*",
        default="CHI_SQUARE",
        help="The statistical association type. Multiple values may be provided "
        "to fall back to. Currently supported values are as follows in order "
        "of time it takes to perform each test: "
        "CHI_SQUARE (fastest), FISHERS, FISHERS_EXACT (slowest)",
        choices=[str(val.name).lower() for val in AssociationTest],
    )
    parser.add_argument(
        "-p",
        "--repeat-for",
        type=str,
        nargs="?",
        help="Perform the association test for every value of the tag.",
    )
