from __future__ import annotations
from argparse import _SubParsersAction, ArgumentParser
from enum import Enum
from os import path
from typing import Iterable
from rich.console import Console
from rich.table import Table
from avariantas.information.mutations import GroupName, Mutation, OrganizedMutations
from avariantas.information.project import Project
from avariantas.ingress.local import (
    load_mutations_from_vcfs,
    load_project,
)
from avariantas.ingress.remote import (
    annotate_mutations_from_ensembl,
)
from avariantas.egress.local import save_project


class ProjectOperations(Enum):
    CREATE = 0
    ADD = 1
    REMOVE = 2
    MUTATIONS = 3
    RESULTS = 4


def add_project_parser(subparsers: _SubParsersAction[ArgumentParser]):
    parser = subparsers.add_parser(
        name="project", help="Manage Avariantas project files."
    )
    parser.add_argument(
        "operation",
        type=str,
        help="Designates what to do with a project.",
        metavar="O",
        choices=[str(val.name).lower() for val in ProjectOperations],
    )
    parser.add_argument(
        "-v",
        "--vcfs",
        type=str,
        help="The directory containing the VCFs to work on. Mutations imported "
        "here will be tagged with origin=<folder_name>. Warning: Folders with "
        "the same name will have the same value for the tag potentially "
        "resulting in confusion while running analysis.",
        nargs="*",
    )
    parser.add_argument(
        "-r", "--recursive", action="store_true", help="Recursively search for VCFs"
    )
    parser.add_argument(
        "-a",
        "--additional-info",
        type=str,
        help="Path to the CSV that describes the sample groups.",
        nargs="?",
        default="",
    )
    parser.add_argument(
        "-additional-info-header",
        type=str,
        help="The header of the column for the VCF filenames in the CSV for the "
        "additional information.",
        nargs="?",
        default="",
    )
    parser.add_argument(
        "-s",
        "--search-1kgenome",
        action="store_true",
        help="Searches the dbSNP for the rsID of all the input mutations for "
        "population sample genotype and allele counts. Adds these as mutation "
        "samples to the working data set of mutations. These will be tagged "
        "with origin=1kgenome",
    )
    parser.add_argument("-p", "--population-code", type=str, nargs="?")


def create_project(
    project_path: str,
    add_local_mutations: Iterable[Mutation] | None,
    search_1k_genome: bool,
    population_code: str,
):
    organized_mutations = OrganizedMutations(
        add_local_mutations if add_local_mutations is not None else []
    )
    if search_1k_genome:
        if add_local_mutations is not None:
            organized_mutations.add_all(
                annotate_mutations_from_ensembl(add_local_mutations, population_code)
            )
        else:
            raise ValueError(
                "Cannot annotate local mutations with 1K Genome as local mutations "
                "were not loaded."
            )
    project = Project(path.basename(project_path), organized_mutations)
    save_project(project, project_path)


def add_mutations_to_project(
    project_path: str,
    add_local_mutations: Iterable[Mutation] | None,
    search_1k_genome: bool,
    population_code: str,
):
    project = load_project(project_path)
    if add_local_mutations is not None:
        project.organized_mutations.add_all(add_local_mutations)
        add_local_mutations = list(add_local_mutations)
    else:
        add_local_mutations = []
    add_local_mutations.extend(project.organized_mutations.mutations)
    if search_1k_genome:
        if add_local_mutations is None:
            raise ValueError("No local mutations to search 1K Genome Database for.")
        project.organized_mutations.add_all(
            annotate_mutations_from_ensembl(add_local_mutations, population_code)
        )
    save_project(project, project_path)


def print_project_mutations_summary(
    project_path: str,
):
    console = Console()
    project = load_project(project_path)
    compound_group_table = Table(show_header=True, header_style="bold green")
    compound_group_table.add_column("Compound Type", style="dim")
    compound_group_table.add_column("Compounded Groups")
    compound_group_table.add_column("Number of Groups")
    compound_group_table.add_column("Number of Mutations")
    for compound in project.organized_mutations.all_compound_groups:
        compound_str_max_len = 10
        compound_str = str(compound)
        if len(compound_str) > compound_str_max_len:
            compound_str = compound_str[: compound_str_max_len - 3] + "..."
        compound_group_table.add_row(
            str(compound.operation),
            compound_str,
            str(len(compound.groups)),
            str(project.organized_mutations.count_occurrences(compound)),
        )
    console.print("[u]Compound Groups[/u]")
    console.print(compound_group_table)
    primitive_group_table = Table(show_header=True, header_style="bold")
    primitive_group_table.add_column("Tags")
    primitive_group_table.add_column("Example Values")
    primitive_group_table.add_column("Total Values")
    primitive_group_table.add_column("Total Mutations")
    for primitive in sorted(project.organized_mutations.known_group_keys):
        example_values = ", ".join(
            [
                val if isinstance(val, str) else str(val)
                for val in project.organized_mutations.known_group_values(primitive)
            ]
        )
        example_max_length = 32
        if len(example_values) > example_max_length:
            example_values = example_values[: example_max_length - 3] + "..."
        primitive_group_table.add_row(
            primitive,
            example_values,
            str(len(project.organized_mutations.known_group_values(primitive))),
            str(
                sum(
                    [
                        len(
                            project.organized_mutations.mutations_for_group(
                                GroupName(primitive, group_val)
                            )
                        )
                        for group_val in project.organized_mutations.known_group_values(
                            primitive
                        )
                    ]
                )
            ),
        )

    console.print("[u]Primitive Groups[/u]")
    console.print(primitive_group_table)


def print_project_results_summary(project_path):
    console = Console()
    project = load_project(project_path)
    results = project.organized_representations
    results_table = Table(show_header=True, header_style="bold green")
    results_table.add_column("Test Type")
    results_table.add_column("Parameters")
    results_table.add_column("Results")
    for rep_name in results.known_representations:
        rep = results.get_representation(rep_name.test_type, rep_name.parameters)
        results_table.add_row(
            rep.test_type,
            ", ".join(str(param) for param in rep.params),
            ", ".join(str(result) for result in rep.results),
        )
    console.print("[u]Results[/u]")
    console.print(results_table)


def manage_project(
    operation_name: str,
    project_path: str,
    vcfs: Iterable[str],
    recursive: bool,
    additional_info: str,
    additional_info_header: str,
    search_1k_genome: bool,
    population_code: str,
):
    operation = ProjectOperations[operation_name.upper()]
    add_local_mutations = None
    if vcfs:
        add_local_mutations = load_mutations_from_vcfs(
            vcf_items=vcfs,
            additional_datatags_csv=additional_info,
            datatags_filename_header=additional_info_header,
            recursive=recursive,
        )

    if operation == ProjectOperations.CREATE:
        create_project(
            project_path, add_local_mutations, search_1k_genome, population_code
        )
    elif operation == ProjectOperations.ADD:
        add_mutations_to_project(
            project_path, add_local_mutations, search_1k_genome, population_code
        )
    elif operation == ProjectOperations.MUTATIONS:
        print_project_mutations_summary(project_path)
    elif operation == ProjectOperations.RESULTS:
        print_project_results_summary(project_path)
